use std::{error::Error, io::Cursor, path::{Path, PathBuf}, process::Command};

const ZIP_BIN: &[u8; 1327616] = include_bytes!("7zip/7za.exe");

fn find_subsequence(haystack: &[u8], needle: &[u8]) -> Vec<usize> {
    haystack.windows(needle.len())
        .enumerate()
        .filter_map(|(i, arr)| if arr == needle { Some(i) } else { None })
        .collect()
}

fn recursively_find_classes(path: PathBuf) -> Vec<PathBuf> {
    let e = std::fs::read_dir(path).unwrap();
    let mut classes = Vec::new();
    for entry in e.flatten() {
        match entry.metadata().unwrap().is_dir() {
            true => classes.append(&mut recursively_find_classes(entry.path())),
            false => {
                if entry.file_name().to_str().unwrap().ends_with(".class") {
                    classes.push(entry.path());
                }
            }
        }
    }
    classes
}
// b"https://sessionserver.mojang.com/session/minecraft/"
fn write_7za() {
    let _ = std::fs::create_dir_all("tmp/");
    let _ = std::fs::write("tmp/7za.exe", ZIP_BIN);
}

pub fn patch_jar(input_jar: &str, output_jar: &str, needles: &[&[u8]], endpoints: &[&str]) -> Result<(), Box<dyn Error + Sync + Send>> {

    write_7za();

    let mut target_dir = PathBuf::new();
    target_dir.push("out");

    let archive = std::fs::read(input_jar)?;

    zip_extract::extract(Cursor::new(archive), &target_dir, true)?;

    for path in recursively_find_classes(PathBuf::from(".\\out")) {
        let mut haystack = std::fs::read(&path).unwrap();
        for index in 0..needles.len() {
            let needle = needles[index];
            let replacement = endpoints[index].as_bytes();
            let mut v = find_subsequence(&haystack, needle);
            if v.is_empty() { continue; }
            while let Some(g) = v.first() {
                let (a,b) = haystack.split_at(*g);
                let l = a[a.len()-1];
                let mut a = a[..a.len()-1].to_vec();
                a.push(  if l as usize > needle.len() { (replacement.len() + (l as usize - needle.len())) as u8 } else { replacement.len() as u8 });
                a.append(&mut replacement.to_vec());
                a.append(&mut b[needle.len()..].to_vec());
                haystack = a;
                v = find_subsequence(&haystack, needle);
            }
        }

        std::fs::write(path, haystack)?;
    }

    zip_dir(&target_dir, output_jar)?;

    std::fs::remove_dir_all(&target_dir)?;

    Ok(())
}


fn zip_dir(
    src_dir: &Path,
    writer: &str,
) -> Result<(), Box<dyn Error + Sync + Send>> {
    let mut a = Command::new(std::fs::canonicalize(PathBuf::from("./tmp/7za.exe")).unwrap())
        .current_dir(src_dir)
        .arg("a")
        .arg("-tzip")
        .arg("-mx=0")
        .arg(writer)
        .arg("*")
        .arg("-r")
        .spawn()?;
    a.wait()?;
    Ok(())
}
